35template <
class T0,
class T1>
39 operator() (T0* d, T1 s)
const noexcept
41 *d =
static_cast<T0
>(s);
45template <
class T0,
class T1>
49 operator() (T0* d, T1 s)
const noexcept
51 *d +=
static_cast<T0
>(s);
55template <
class T0,
class T1>
61 operator() (U0* d, T1 s)
const noexcept
63 Gpu::Atomic::AddNoRet(d,
static_cast<U0
>(s));
67template <
class T0,
class T1,
class F>
69fab_to_fab (Vector<Array4CopyTag<T0, T1> >
const& copy_tags,
int scomp,
int dcomp,
int ncomp,
72 TagVector<Array4CopyTag<T0, T1>> tv{copy_tags};
74 detail::ParallelFor_doit(tv,
76 int icell,
int ncells,
int i,
int j,
int k, Array4CopyTag<T0, T1>
const& tag)
noexcept
79 for (
int n = 0; n < ncomp; ++n) {
80 f(&(tag.dfab(i,j,k,n+dcomp)),
81 tag.sfab(i+tag.offset.x,j+tag.offset.y,k+tag.offset.z,n+scomp));
87template <
class TagType,
class F>
89fab_to_fab_store (Vector<TagType>
const& tags,
int scomp,
int dcomp,
int ncomp,
F&&f)
96 for (
int n = 0; n < ncomp; ++n) {
97 f(&(tag.dfab(i,j,k,n+dcomp)),
98 tag.sfab(i+tag.offset.x,j+tag.offset.y,k+tag.offset.z,n+scomp));
104template <
class TagType,
class F>
106fab_to_fab_other (Vector<TagType>
const& tags,
int scomp,
int dcomp,
int ncomp,
F&&f)
111 int* m = &(tag.mask(i,j,k));
112 bool my_turn =
false;
114#if defined(AMREX_USE_SYCL)
120#if defined(AMREX_USE_SYCL)
121 sycl::atomic_fence(sycl::memory_order::acq_rel, sycl::memory_scope::device);
125 for (
int n = 0; n < ncomp; ++n) {
126 f(&(tag.dfab(i,j,k,n+dcomp)),
127 tag.sfab(i+tag.offset.x,j+tag.offset.y,k+tag.offset.z,n+scomp));
129#if defined(AMREX_USE_SYCL)
130 sycl::atomic_fence(sycl::memory_order::acq_rel, sycl::memory_scope::device);
137#if defined(AMREX_USE_CUDA)
139#if defined(__CUDA_ARCH__) && (__CUDA_ARCH__ < 700)
142 for (
int c = 0; c < 2; ++c) {
146 for (
int c = 0; c < 2; ++c) {
147 __asm__
volatile(
"");
154#elif defined(AMREX_USE_HIP)
156 __builtin_amdgcn_s_sleep(1);
158#elif defined(AMREX_USE_SYCL)
160 for (
int c = 0; c < 2; ++c) {
161 __asm__
volatile(
"");
170template <
class T0,
class T1,
class F>
172fab_to_fab (Vector<Array4CopyTag<T0, T1> >
const& copy_tags,
int scomp,
int dcomp,
173 int ncomp,
F && f, Vector<Array4Tag<int> >
const& masks)
175 using TagType = Array4MaskCopyTag<T0, T1>;
176 Vector<TagType> tags;
177 const int N = copy_tags.size();
179 for (
int i = 0; i < N; ++i) {
180 tags.push_back(TagType{.dfab = copy_tags[i].dfab, .sfab = copy_tags[i].sfab,
181 .mask = masks[i].dfab, .dbox = copy_tags[i].dbox,
182 .offset = copy_tags[i].offset});
185 if constexpr (std::is_same_v<F, CellStore<T0,T1>>)
187 fab_to_fab_store(tags, scomp, dcomp, ncomp, std::forward<F>(f));
191 fab_to_fab_other(tags, scomp, dcomp, ncomp, std::forward<F>(f));
196template <
typename T0,
typename T1>
199fab_to_fab_atomic_cpy (Vector<Array4CopyTag<T0, T1> >
const& copy_tags,
int scomp,
200 int dcomp,
int ncomp, Vector<Array4Tag<int> >
const&)
202 fab_to_fab<T0, T1>(copy_tags, scomp, dcomp, ncomp, CellStore<T0, T1>());
205template <
typename T0,
typename T1>
208fab_to_fab_atomic_cpy (Vector<Array4CopyTag<T0, T1> >
const& copy_tags,
int scomp,
209 int dcomp,
int ncomp, Vector<Array4Tag<int> >
const& masks)
211 fab_to_fab(copy_tags, scomp, dcomp, ncomp, CellStore<T0, T1>(), masks);
214template <
typename T0,
typename T1>
217fab_to_fab_atomic_add (Vector<Array4CopyTag<T0, T1> >
const& copy_tags,
int scomp,
218 int dcomp,
int ncomp, Vector<Array4Tag<int> >
const&)
220 fab_to_fab(copy_tags, scomp, dcomp, ncomp, CellAtomicAdd<T0, T1>());
223template <
typename T0,
typename T1>
226fab_to_fab_atomic_add (Vector<Array4CopyTag<T0, T1> >
const& copy_tags,
int scomp,
227 int dcomp,
int ncomp, Vector<Array4Tag<int> >
const& masks)
229 fab_to_fab(copy_tags, scomp, dcomp, ncomp, CellAdd<T0, T1>(), masks);
232template <
typename T0,
typename T1,
class F>
233void deterministic_fab_to_fab (Vector<Array4CopyTag<T0,T1>>
const& a_tags,
int scomp,
234 int dcomp,
int ncomp,
F const& f)
236 if (a_tags.empty()) {
return; }
238 using TagType = Array4CopyTag<T0,T1>;
242 std::pair<int,Box> dindex_tilebox;
243 bool operator< (TiledTag
const& rhs)
const noexcept {
244 return this->dindex_tilebox < rhs.dindex_tilebox;
246 bool operator!= (TiledTag
const& rhs)
const noexcept {
247 return this->dindex_tilebox != rhs.dindex_tilebox;
250 Vector<TiledTag> tiled_tags;
252 auto const ixtype = a_tags[0].dbox.ixType();
254 constexpr int tile_size = 64;
255 for (
int itag = 0; itag < a_tags.size(); ++itag) {
256 auto const& tag = a_tags[itag];
257 auto const& dlo = tag.dbox.smallEnd();
258 auto const& dhi = tag.dbox.bigEnd();
260 amrex::coarsen<tile_size>(dlo[1]),
261 amrex::coarsen<tile_size>(dlo[2])));
263 amrex::coarsen<tile_size>(dhi[1]),
264 amrex::coarsen<tile_size>(dhi[2])));
265#if (AMREX_SPACEDIM == 3)
266 for (
int kt = tlo[2]; kt <= thi[2]; ++kt)
269#if (AMREX_SPACEDIM >= 2)
270 for (
int jt = tlo[1]; jt <= thi[1]; ++jt)
273 for (
int it = tlo[0]; it <= thi[0]; ++it)
278 tiled_tags.push_back(TiledTag{
280 .dindex_tilebox = std::make_pair(tag.dindex,
Box(lo, lo+(tile_size-1), ixtype))
287 std::sort(tiled_tags.begin(), tiled_tags.end());
289 Gpu::HostVector<unsigned int> h_ntags;
290 Gpu::HostVector<TagType> h_tags;
291 h_tags.reserve(tiled_tags.size());
293 for (
unsigned int itag = 0; itag < tiled_tags.size(); ++itag) {
295 h_ntags.push_back(0);
296 }
else if (tiled_tags[itag-1] != tiled_tags[itag]) {
297 h_ntags.push_back(itag);
299 auto const& ttag = tiled_tags[itag];
300 auto const& btag = a_tags[ttag.tag_index];
301 h_tags.push_back(TagType{.dfab = btag.dfab, .dindex = btag.dindex, .sfab = btag.sfab,
302 .dbox = btag.dbox & ttag.dindex_tilebox.second,
303 .offset = btag.offset});
305 h_ntags.push_back((
unsigned int)tiled_tags.size());
307 Gpu::DeviceVector<TagType> d_tags(h_tags.size());
308 Gpu::DeviceVector<unsigned int> d_ntags(h_ntags.size());
311 auto const* ptag = d_tags.data();
312 auto const* pntags = d_ntags.data();
313 auto const nblocks =
int(h_ntags.size()-1);
314 constexpr auto nthreads = 256;
318 [[sycl::reqd_work_group_size(nthreads)]]
324 Dim1 blockIdx{item.get_group_linear_id()};
325 Dim1 threadIdx{item.get_local_linear_id()};
328 for (
unsigned int itag = pntags[blockIdx.x]; itag < pntags[blockIdx.x+1]; ++itag) {
329 auto const tag = ptag[itag];
330 auto ncells =
int(tag.dbox.numPts());
333 for (
int icell =
int(threadIdx.x); icell < ncells; icell += nthreads) {
334 int k = icell / (len.x*len.y);
335 int j = (icell - k*(len.x*len.y)) / len.x;
336 int i = (icell - k*(len.x*len.y)) - j*len.x;
340 for (
int n = 0; n < ncomp; ++n) {
341 f(tag.dfab.ptr(i,j,k,n+dcomp),
342 tag.sfab(i + tag.offset.x,
344 k + tag.offset.z, n+scomp));
348 if (itag+1 < pntags[blockIdx.x+1]) {
350 sycl::group_barrier(item.get_group());
360template <
typename B,
typename V,
typename TT>
362void unpack_recv_buffer_gpu_atomic_add (
char* pbuffer, TagVector<TT>
const& tv,
363 int dcomp,
int ncomp)
365 detail::ParallelFor_doit(tv,
367 int icell,
int ncells,
int i,
int j,
int k, TT
const& tag)
noexcept
369 if (icell < ncells) {
370 Array4<B const> sfab{(B
const*)(pbuffer+tag.poff),
372 for (
int n = 0; n < ncomp; ++n) {
380template <
typename B,
typename V,
typename TT>
382void unpack_recv_buffer_gpu_atomic_add (
char* pbuffer, TagVector<TT>
const& tv,
383 int dcomp,
int ncomp)
386 amrex::Abort(
"unpack_recv_buffer_gpu: should NOT get here");
398 auto const& LocTags = *(TheFB.
m_LocTags);
399 auto N_locs =
static_cast<int>(LocTags.size());
400 if (N_locs == 0) {
return; }
405#pragma omp parallel for
407 for (
int i = 0; i < N_locs; ++i)
416 dfab->template copy<RunOn::Host>(*sfab, tag.
sbox, scomp, tag.
dbox, scomp, ncomp);
422 for (
int i = 0; i < N_locs; ++i)
429 loc_copy_tags[tag.
dstIndex].push_back
437 const auto& tags = loc_copy_tags[mfi];
438 auto dfab = this->array(mfi);
439 for (
auto const & tag : tags)
441 auto const sfab = tag.sfab->array();
442 const auto offset = tag.offset.dim3();
444 [=] (
int i,
int j,
int k,
int n)
noexcept
446 dfab(i,j,k,n+scomp) = sfab(i+offset.x,j+offset.y,k+offset.z,n+scomp);
457 auto const& LocTags = *(TheFB.
m_LocTags);
458 auto N_locs =
static_cast<int>(LocTags.size());
459 if (N_locs == 0) {
return; }
463 std::vector<FAB> src_fabs(N_locs);
464 for (
int itag = 0; itag < N_locs; ++itag) {
466 src_fabs[itag].resize(tag.
sbox,ncomp);
467 loc_copy_tags[tag.
dstIndex].push_back
473#pragma omp parallel for
475 for (
int itag = 0; itag < N_locs; ++itag) {
477 src_fabs[itag].template copy<RunOn::Host>(this->
operator[](tag.
srcIndex), scomp, 0, ncomp);
485 const auto& tags = loc_copy_tags[mfi];
486 const auto& dfab = this->array(mfi);
487 for (
auto const & tag : tags)
489 auto const sfab = tag.sfab->array();
490 const auto offset = tag.offset.dim3();
492 [&] (
int i,
int j,
int k,
int n)
noexcept
494 dfab(i,j,k,n+scomp) += sfab(i+offset.x,j+offset.y,k+offset.z,n);
506 auto const& LocTags = *(TheFB.m_LocTags);
507 int N_locs = LocTags.size();
512 if (
auto it = m_fb_local_copy_handler.find(TheFB.m_id);
513 it != m_fb_local_copy_handler.end())
515 tv = it->second.get();
518 loc_copy_tags.reserve(N_locs);
520 for (
int i = 0; i < N_locs; ++i)
527 int li = this->localindex(tag.
dstIndex);
528 loc_copy_tags.push_back
529 (TagType{.dfab = this->atLocalIdx(li).array(),
531 .sfab = this->fabPtr(tag.
srcIndex)->const_array(),
536 auto utv = std::make_unique<TagVector<TagType>>(loc_copy_tags);
538 m_fb_local_copy_handler[TheFB.m_id] = std::move(utv);
548 auto const& LocTags = *(TheFB.
m_LocTags);
549 int N_locs = LocTags.size();
550 if (N_locs == 0) {
return; }
557 auto* tv = FB_get_local_copy_tag_vector(TheFB);
559 detail::ParallelFor_doit(*tv,
561 int icell,
int ncells,
int i,
int j,
int k, TagType
const& tag)
noexcept
563 if (icell < ncells) {
564 for (
int n = 0; n < ncomp; ++n) {
565 tag.dfab(i,j,k,n+scomp) = tag.sfab(i+tag.offset.x,
567 k+tag.offset.z,n+scomp);
575 loc_copy_tags.reserve(N_locs);
579 masks_unique.reserve(this->local_size());
581 masks.reserve(N_locs);
583 for (
int i = 0; i < N_locs; ++i)
590 int li = this->localindex(tag.
dstIndex);
591 loc_copy_tags.push_back
592 (TagType{.dfab = this->atLocalIdx(li).array(),
594 .sfab = this->fabPtr(tag.
srcIndex)->const_array(),
598 if (!maskfabs[li].isAllocated()) {
599 maskfabs[li].resize(this->atLocalIdx(li).box());
611 detail::fab_to_fab_atomic_cpy<value_type, value_type>(
612 loc_copy_tags, scomp, scomp, ncomp, masks);
620 auto const& LocTags = *(TheFB.
m_LocTags);
621 int N_locs = LocTags.size();
622 if (N_locs == 0) {
return; }
627 loc_copy_tags_1.reserve(N_locs);
628 loc_copy_tags_2.reserve(N_locs);
631 for (
int itag = 0; itag < N_locs; ++itag) {
633 src_fabs[itag].resize(tag.
sbox,ncomp);
634 loc_copy_tags_1.push_back(
635 TagType{.dfab = src_fabs[itag].array(), .dindex = -1,
636 .sfab = this->const_array(tag.
srcIndex,scomp), .dbox = tag.
sbox,
637 .offset =
Dim3{.x = 0, .y = 0, .z = 0}});
638 loc_copy_tags_2.push_back(
640 .sfab = src_fabs[itag].const_array(), .dbox = tag.
dbox,
648 detail::fab_to_fab(loc_copy_tags_1, 0, 0, ncomp,
649 detail::CellStore<value_type, value_type>{});
651 detail::deterministic_fab_to_fab(loc_copy_tags_2, 0, 0, ncomp,
652 detail::CellAdd<value_type,value_type>{});
655 detail::fab_to_fab(loc_copy_tags_2, 0, 0, ncomp,
656 detail::CellAtomicAdd<value_type, value_type>{});
669 auto const& LocTags = *(thecmd.
m_LocTags);
670 int N_locs = LocTags.
size();
671 if (N_locs == 0) {
return; }
676 loc_setval_tags.reserve(N_locs);
680 for (
int i = 0; i < N_locs; ++i)
684 loc_setval_tags.push_back(TagType{.dfab = this->array(tag.
dstIndex), .dbox = tag.
dbox});
688 [
x,scomp]
AMREX_GPU_DEVICE (
int i,
int j,
int k,
int n, TagType
const& tag)
noexcept
690 tag.dfab(i,j,k,n+scomp) =
x;
699 auto const& RcvTags = *(thecmd.
m_RcvTags);
705 for (
auto it = RcvTags.begin(); it != RcvTags.end(); ++it) {
706 for (
auto const& tag: it->second) {
707 rcv_setval_tags.push_back(TagType{.dfab = this->array(tag.dstIndex), .dbox = tag.dbox});
711 if (rcv_setval_tags.empty()) {
return; }
716 [
x,scomp]
AMREX_GPU_DEVICE (
int i,
int j,
int k,
int n, TagType
const& tag)
noexcept
718 tag.dfab(i,j,k,n+scomp) =
x;
722#if defined(__CUDACC__) && defined (AMREX_USE_CUDA)
727 const int N_locs = (*TheFB.m_LocTags).size();
729 for (
int i = 0; i < N_locs; ++i)
731 const CopyComTag& tag = (*TheFB.m_LocTags)[i];
736 loc_copy_tags[tag.dstIndex].push_back
737 ({this->fabPtr(tag.srcIndex), tag.dbox, tag.sbox.smallEnd()-tag.dbox.smallEnd()});
741 if ( !(TheFB.m_localCopy.ready()) )
743 const_cast<FB&
>(TheFB).m_localCopy.resize(N_locs);
747 for (MFIter mfi(*
this, MFItInfo().DisableDeviceSync()); mfi.isValid(); ++mfi)
749 amrex::Gpu::Device::startGraphRecording( (mfi.LocalIndex() == 0),
750 const_cast<FB&
>(TheFB).m_localCopy.getHostPtr(0),
751 (TheFB).m_localCopy.getDevicePtr(0),
752 std::size_t(
sizeof(CopyMemory)*N_locs) );
754 const auto& tags = loc_copy_tags[mfi];
755 for (
auto const & tag : tags)
757 const auto offset = tag.offset.dim3();
758 CopyMemory* cmem = TheFB.m_localCopy.getDevicePtr(idx++);
762 auto const dst = cmem->getDst<value_type>();
763 auto const src = cmem->getSrc<value_type>();
764 for (int n = 0; n < cmem->ncomp; ++n) {
765 dst(i,j,k,(cmem->scomp)+n) = src(i+offset.x,j+offset.y,k+offset.z,(cmem->scomp)+n);
770 bool last_iter = mfi.LocalIndex() == (this->local_size()-1);
771 cudaGraphExec_t graphExec = amrex::Gpu::Device::stopGraphRecording(last_iter);
772 if (last_iter) {
const_cast<FB&
>(TheFB).m_localCopy.setGraph( graphExec ); }
781 for (MFIter mfi(*
this); mfi.isValid(); ++mfi)
783 auto const dst_array = this->array(mfi);
784 const auto& tags = loc_copy_tags[mfi];
785 for (
auto const & tag : tags)
787 const_cast<FB&
>(TheFB).m_localCopy.setParams(idx++, makeCopyMemory(tag.sfab->array(),
794 TheFB.m_localCopy.executeGraph();
800FabArray<FAB>::FB_local_copy_cuda_graph_n (
const FB& TheFB,
int scomp,
int ncomp)
802 const int N_locs = TheFB.m_LocTags->size();
806 for (
int i = 0; i < N_locs; ++i)
808 const CopyComTag& tag = (*TheFB.m_LocTags)[i];
810 BL_ASSERT(ParallelDescriptor::sameTeam(distributionMap[tag.dstIndex]));
811 BL_ASSERT(ParallelDescriptor::sameTeam(distributionMap[tag.srcIndex]));
813 if (distributionMap[tag.dstIndex] == ParallelDescriptor::MyProc())
815 loc_copy_tags[tag.dstIndex].push_back
816 ({this->fabPtr(tag.srcIndex), tag.dbox, tag.sbox.smallEnd()-tag.dbox.smallEnd()});
823 if ( !(TheFB.m_localCopy.ready()) )
825 const_cast<FB&
>(TheFB).m_localCopy.resize(launches);
829 for (MFIter mfi(*
this, MFItInfo().DisableDeviceSync()); mfi.isValid(); ++mfi)
831 const auto& tags = loc_copy_tags[mfi];
832 for (
int t = 0; t<tags.size(); ++t)
834 Gpu::Device::setStreamIndex(cuda_stream++);
835 amrex::Gpu::Device::startGraphRecording( (idx == 0),
836 const_cast<FB&
>(TheFB).m_localCopy.getHostPtr(0),
837 (TheFB).m_localCopy.getDevicePtr(0),
838 std::size_t(
sizeof(CopyMemory)*launches) );
840 const auto& tag = tags[t];
841 const Dim3
offset = tag.offset.dim3();
843 CopyMemory* cmem = TheFB.m_localCopy.getDevicePtr(idx++);
846 auto const dst = cmem->getDst<value_type>();
847 auto const src = cmem->getSrc<value_type>();
848 for (int n = 0; n < cmem->ncomp; ++n) {
849 dst(i,j,k,(cmem->scomp)+n) = src(i+offset.x,j+offset.y,k+offset.z,(cmem->scomp)+n);
853 bool last_iter = idx == launches;
854 cudaGraphExec_t graphExec = Gpu::Device::stopGraphRecording(last_iter);
855 if (last_iter) {
const_cast<FB&
>(TheFB).m_localCopy.setGraph( graphExec ); }
863 for (MFIter mfi(*
this); mfi.isValid(); ++mfi)
865 const auto& dst_array = this->array(mfi);
866 const auto& tags = loc_copy_tags[mfi];
867 for (
auto const & tag : tags)
869 const_cast<FB&
>(TheFB).m_localCopy.setParams(idx++, makeCopyMemory(tag.sfab->array(),
876 TheFB.m_localCopy.executeGraph(
false);
888#if defined(__CUDACC__) && defined(AMREX_USE_CUDA)
892FabArray<FAB>::FB_pack_send_buffer_cuda_graph (
const FB& TheFB,
int scomp,
int ncomp,
893 Vector<char*>& send_data,
894 Vector<std::size_t>
const& send_size,
895 Vector<
typename FabArray<FAB>::CopyComTagsContainer
const*>
const& send_cctc)
897 const int N_snds = send_data.size();
898 if (N_snds == 0) {
return; }
900 if ( !(TheFB.m_copyToBuffer.ready()) )
905 for (
int send = 0; send < N_snds; ++send) {
906 if (send_size[send] > 0) {
907 launches += send_cctc[send]->size();
910 const_cast<FB&
>(TheFB).m_copyToBuffer.resize(launches);
914 for (Gpu::StreamIter sit(N_snds,Gpu::StreamItInfo().DisableDeviceSync());
915 sit.isValid(); ++sit)
917 amrex::Gpu::Device::startGraphRecording( (sit() == 0),
918 const_cast<FB&
>(TheFB).m_copyToBuffer.getHostPtr(0),
919 (TheFB).m_copyToBuffer.getDevicePtr(0),
920 std::size_t(
sizeof(CopyMemory)*launches) );
923 if (send_size[j] > 0)
925 auto const& cctc = *send_cctc[j];
926 for (
auto const& tag : cctc)
928 const Box& bx = tag.sbox;
929 CopyMemory* cmem = TheFB.m_copyToBuffer.getDevicePtr(idx++);
932 auto const pfab = cmem->getDst<value_type>();
933 auto const sfab = cmem->getSrc<value_type>();
934 for (
int n = 0; n < cmem->ncomp; ++n)
936 pfab(ii,jj,kk,n) = sfab(ii,jj,kk,n+(cmem->scomp));
942 bool last_iter = sit() == (N_snds-1);
943 cudaGraphExec_t graphExec = amrex::Gpu::Device::stopGraphRecording(last_iter);
944 if (last_iter) {
const_cast<FB&
>(TheFB).m_copyToBuffer.setGraph( graphExec ); }
950 for (
int send = 0; send < N_snds; ++send)
953 if (send_size[j] > 0)
955 char* dptr = send_data[j];
956 auto const& cctc = *send_cctc[j];
957 for (
auto const& tag : cctc)
959 const_cast<FB&
>(TheFB).m_copyToBuffer.setParams(idx++, makeCopyMemory(this->array(tag.srcIndex),
965 dptr += (tag.sbox.numPts() * ncomp *
sizeof(value_type));
968 BL_ASSERT(dptr <= send_data[j] + send_size[j]);
973 TheFB.m_copyToBuffer.executeGraph();
978FabArray<FAB>::FB_unpack_recv_buffer_cuda_graph (
const FB& TheFB,
int dcomp,
int ncomp,
979 Vector<char*>
const& recv_data,
980 Vector<std::size_t>
const& recv_size,
981 Vector<CopyComTagsContainer const*>
const& recv_cctc,
984 const int N_rcvs = recv_cctc.size();
985 if (N_rcvs == 0) {
return; }
989 for (
int k = 0; k < N_rcvs; ++k)
991 if (recv_size[k] > 0)
993 const char* dptr = recv_data[k];
994 auto const& cctc = *recv_cctc[k];
995 for (
auto const& tag : cctc)
997 recv_copy_tags[tag.dstIndex].push_back(VoidCopyTag{.p = dptr, .dbox = tag.dbox});
998 dptr += tag.dbox.numPts() * ncomp *
sizeof(value_type);
1002 BL_ASSERT(dptr <= recv_data[k] + recv_size[k]);
1006 if ( !(TheFB.m_copyFromBuffer.ready()) )
1008 const_cast<FB&
>(TheFB).m_copyFromBuffer.resize(launches);
1011 for (MFIter mfi(*
this, MFItInfo().DisableDeviceSync()); mfi.isValid(); ++mfi)
1013 amrex::Gpu::Device::startGraphRecording( (mfi.LocalIndex() == 0),
1014 const_cast<FB&
>(TheFB).m_copyFromBuffer.getHostPtr(0),
1015 (TheFB).m_copyFromBuffer.getDevicePtr(0),
1016 std::size_t(
sizeof(CopyMemory)*launches) );
1018 const auto& tags = recv_copy_tags[mfi];
1019 for (
auto const & tag : tags)
1021 CopyMemory* cmem = TheFB.m_copyFromBuffer.getDevicePtr(idx++);
1024 auto const pfab = cmem->getSrc<value_type>();
1025 auto const dfab = cmem->getDst<value_type>();
1026 for (int n = 0; n < cmem->ncomp; ++n)
1028 dfab(i,j,k,n+(cmem->scomp)) = pfab(i,j,k,n);
1033 bool last_iter = mfi.LocalIndex() == (this->local_size()-1);
1034 cudaGraphExec_t graphExec = amrex::Gpu::Device::stopGraphRecording(last_iter);
1035 if (last_iter) {
const_cast<FB&
>(TheFB).m_copyFromBuffer.setGraph( graphExec ); }
1041 for (MFIter mfi(*
this); mfi.isValid(); ++mfi)
1043 auto dst_array = this->array(mfi);
1044 const auto & tags = recv_copy_tags[mfi];
1045 for (
auto const & tag : tags)
1047 const_cast<FB&
>(TheFB).m_copyFromBuffer.setParams(idx++, makeCopyMemory(
amrex::makeArray4((value_type*)(tag.p),
1056 TheFB.m_copyFromBuffer.executeGraph();
1062template <
typename BUF>
1067 int ncomp, std::uint64_t
id)
const
1072 auto kit = std::find_if(send_cctc.begin(), send_cctc.end(),
1074 if (kit == send_cctc.end()) {
1081 char* pbuf = send_data[0];
1082 const int N_snds = send_data.
size();
1083 for (
int j = 0; j < N_snds; ++j)
1085 if (send_size[j] > 0)
1087 char* dptr = send_data[j];
1088 auto const& cctc = *send_cctc[j];
1089 for (
auto const& tag : cctc)
1091 snd_copy_tags.emplace_back
1092 (TagType{.sfab = this->const_array(tag.srcIndex), .poff = dptr-pbuf, .bx = tag.sbox});
1093 dptr += (tag.sbox.numPts() * ncomp *
sizeof(BUF));
1097 return snd_copy_tags;
1101 std::tuple<std::uint64_t,std::size_t,int> key{id,
sizeof(BUF), ncomp};
1103 if (
auto it = m_send_copy_handler.find(key); it != m_send_copy_handler.end()) {
1104 tv = it->second.get();
1106 if (m_send_copy_handler.size() > 32) {
1111 m_send_copy_handler.clear();
1113 auto snd_copy_tags = get_tags();
1114 auto utv = std::make_unique<TagVector<TagType>>(snd_copy_tags);
1116 m_send_copy_handler[key] = std::move(utv);
1123template <
typename BUF>
1131 const int N_snds = send_data.
size();
1132 if (N_snds == 0) {
return; }
1136 auto* tv = src.template get_send_copy_tag_vector<BUF>
1137 (send_data, send_size, send_cctc, ncomp,
id);
1138 if (tv ==
nullptr) {
return; }
1140 char* pbuffer = send_data[0];
1142 detail::ParallelFor_doit(*tv,
1144 int icell,
int ncells,
int i,
int j,
int k, TagType
const& tag)
noexcept
1146 if (icell < ncells) {
1149 for (
int n = 0; n < ncomp; ++n) {
1150 dfab(i,j,k,n) = (BUF)tag.sfab(i,j,k,n+scomp);
1155 Gpu::streamSynchronize();
1159template <
typename BUF>
1164 int ncomp, std::uint64_t
id)
1169 auto kit = std::find_if(recv_cctc.begin(), recv_cctc.end(),
1171 if (kit == recv_cctc.end()) {
1178 char* pbuf = recv_data[0];
1179 const int N_rcvs = recv_cctc.
size();
1180 for (
int k = 0; k < N_rcvs; ++k)
1182 if (recv_size[k] > 0)
1184 char* dptr = recv_data[k];
1185 auto const& cctc = *recv_cctc[k];
1186 for (
auto const& tag : cctc)
1188 const int li = this->localindex(tag.dstIndex);
1189 recv_copy_tags.emplace_back
1190 (TagType{.dfab = this->atLocalIdx(li).array(), .poff = dptr-pbuf, .bx = tag.dbox});
1191 dptr += tag.dbox.numPts() * ncomp *
sizeof(BUF);
1195 return recv_copy_tags;
1199 std::tuple<std::uint64_t,std::size_t,int> key{id,
sizeof(BUF), ncomp};
1201 if (
auto it = m_recv_copy_handler.find(key); it != m_recv_copy_handler.end()) {
1202 tv = it->second.get();
1204 if (m_recv_copy_handler.size() > 32) {
1209 m_recv_copy_handler.clear();
1211 auto recv_copy_tags = get_tags();
1212 auto utv = std::make_unique<TagVector<TagType>>(recv_copy_tags);
1214 m_recv_copy_handler[key] = std::move(utv);
1221template <
typename BUF>
1227 CpOp op,
bool is_thread_safe, std::uint64_t
id,
1230 const int N_rcvs = recv_cctc.
size();
1231 if (N_rcvs == 0) {
return; }
1233 bool use_mask =
false;
1234 if (!is_thread_safe)
1237 (op == FabArrayBase::ADD && !amrex::HasAtomicAdd <value_type>::value))
1248 tags.reserve(N_rcvs);
1249 for (
int k = 0; k < N_rcvs; ++k) {
1250 if (recv_size[k] > 0) {
1251 char const* dptr = recv_data[k];
1252 auto const& cctc = *recv_cctc[k];
1253 for (
auto const& tag : cctc) {
1255 TagType{.dfab = dst.
array(tag.dstIndex), .dindex = tag.dstIndex,
1260 .offset =
Dim3{.
x = 0, .y = 0, .z = 0}});
1261 dptr += tag.dbox.numPts() * ncomp *
sizeof(BUF);
1266 detail::deterministic_fab_to_fab<value_type,BUF>
1267 (tags, 0, dcomp, ncomp, detail::CellAdd<value_type,BUF>{});
1275 auto* tv = dst.template get_recv_copy_tag_vector<BUF>
1276 (recv_data, recv_size, recv_cctc, ncomp,
id);
1277 if (tv ==
nullptr) {
return; }
1279 char* pbuffer = recv_data[0];
1281 if (op == FabArrayBase::COPY)
1283 detail::ParallelFor_doit(*tv,
1285 int icell,
int ncells,
int i,
int j,
int k, TagType
const& tag)
noexcept
1287 if (icell < ncells) {
1290 for (
int n = 0; n < ncomp; ++n) {
1291 tag.dfab(i,j,k,n+dcomp) = (
value_type)sfab(i,j,k,n);
1298 if (is_thread_safe) {
1299 detail::ParallelFor_doit(*tv,
1301 int icell,
int ncells,
int i,
int j,
int k, TagType
const& tag)
noexcept
1303 if (icell < ncells) {
1306 for (
int n = 0; n < ncomp; ++n) {
1307 tag.dfab(i,j,k,n+dcomp) += (
value_type)sfab(i,j,k,n);
1312 detail::unpack_recv_buffer_gpu_atomic_add<BUF, value_type>
1313 (pbuffer, *tv, dcomp, ncomp);
1316 Gpu::streamSynchronize();
1320 char* pbuffer = recv_data[0];
1324 recv_copy_tags.reserve(N_rcvs);
1331 for (
int k = 0; k < N_rcvs; ++k)
1333 if (recv_size[k] > 0)
1335 std::size_t
offset = recv_data[k]-recv_data[0];
1336 const char* dptr = pbuffer +
offset;
1337 auto const& cctc = *recv_cctc[k];
1338 for (
auto const& tag : cctc)
1341 recv_copy_tags.emplace_back(TagType{
1342 .dfab = dst.
atLocalIdx(li).array(), .dindex = tag.dstIndex,
1345 .offset =
Dim3{.
x = 0, .y = 0, .z = 0}
1347 dptr += tag.dbox.numPts() * ncomp *
sizeof(BUF);
1349 if (!maskfabs[li].isAllocated()) {
1350 maskfabs[li].resize(dst.
atLocalIdx(li).box());
1362 msk.
dfab(i,j,k) = 0;
1365 if (op == FabArrayBase::COPY)
1367 detail::fab_to_fab_atomic_cpy<value_type, BUF>(
1368 recv_copy_tags, 0, dcomp, ncomp, masks);
1372 detail::fab_to_fab_atomic_add<value_type, BUF>(
1373 recv_copy_tags, 0, dcomp, ncomp, masks);
1383template <
typename BUF>
1392 auto const N_snds =
static_cast<int>(send_data.
size());
1393 if (N_snds == 0) {
return; }
1396#pragma omp parallel for
1398 for (
int j = 0; j < N_snds; ++j)
1400 if (send_size[j] > 0)
1402 char* dptr = send_data[j];
1403 auto const& cctc = *send_cctc[j];
1404 for (
auto const& tag : cctc)
1406 const Box& bx = tag.sbox;
1407 auto const sfab = src.
array(tag.srcIndex);
1410 [=] (
int ii,
int jj,
int kk,
int n)
noexcept
1412 pfab(ii,jj,kk,n) =
static_cast<BUF
>(sfab(ii,jj,kk,n+scomp));
1414 dptr += (bx.
numPts() * ncomp *
sizeof(BUF));
1416 BL_ASSERT(dptr <= send_data[j] + send_size[j]);
1422template <
typename BUF>
1428 CpOp op,
bool is_thread_safe)
1432 auto const N_rcvs =
static_cast<int>(recv_cctc.
size());
1433 if (N_rcvs == 0) {
return; }
1438#pragma omp parallel for
1440 for (
int k = 0; k < N_rcvs; ++k)
1442 if (recv_size[k] > 0)
1444 const char* dptr = recv_data[k];
1445 auto const& cctc = *recv_cctc[k];
1446 for (
auto const& tag : cctc)
1448 const Box& bx = tag.dbox;
1449 FAB& dfab = dst[tag.dstIndex];
1450 if (op == FabArrayBase::COPY)
1452 dfab.template copyFromMem<RunOn::Host, BUF>(bx, dcomp, ncomp, dptr);
1456 dfab.template addFromMem<RunOn::Host, BUF>(tag.dbox, dcomp, ncomp, dptr);
1458 dptr += bx.
numPts() * ncomp *
sizeof(BUF);
1460 BL_ASSERT(dptr <= recv_data[k] + recv_size[k]);
1466 LayoutData<Vector<VoidCopyTag> > recv_copy_tags;
1468 for (
int k = 0; k < N_rcvs; ++k)
1470 if (recv_size[k] > 0)
1472 const char* dptr = recv_data[k];
1473 auto const& cctc = *recv_cctc[k];
1474 for (
auto const& tag : cctc)
1476 recv_copy_tags[tag.dstIndex].push_back(VoidCopyTag{.p = dptr, .dbox = tag.dbox});
1477 dptr += tag.dbox.numPts() * ncomp *
sizeof(BUF);
1479 BL_ASSERT(dptr <= recv_data[k] + recv_size[k]);
1486 for (MFIter mfi(dst); mfi.isValid(); ++mfi)
1488 const auto& tags = recv_copy_tags[mfi];
1489 auto dfab = dst.
array(mfi);
1490 for (
auto const & tag : tags)
1493 if (op == FabArrayBase::COPY)
1496 [=] (
int i,
int j,
int k,
int n)
noexcept
1498 dfab(i,j,k,n+dcomp) = pfab(i,j,k,n);
1504 [=] (
int i,
int j,
int k,
int n)
noexcept
1506 dfab(i,j,k,n+dcomp) += pfab(i,j,k,n);
#define BL_ASSERT(EX)
Definition AMReX_BLassert.H:39
#define AMREX_ALWAYS_ASSERT(EX)
Definition AMReX_BLassert.H:50
#define AMREX_FORCE_INLINE
Definition AMReX_Extension.H:124
#define AMREX_HOST_DEVICE_FOR_3D(...)
Definition AMReX_GpuLaunchMacrosC.nolint.H:106
#define AMREX_GPU_DEVICE
Definition AMReX_GpuQualifiers.H:18
Array4< int const > offset
Definition AMReX_HypreMLABecLap.cpp:1129
#define AMREX_D_DECL(a, b, c)
Definition AMReX_SPACE.H:171
__host__ __device__ Long numPts() const noexcept
Return the number of points contained in the BoxND.
Definition AMReX_Box.H:385
__host__ __device__ const IntVectND< dim > & smallEnd() const &noexcept
Return the inclusive lower bound of the box.
Definition AMReX_Box.H:124
CopyComTag::CopyComTagsContainer CopyComTagsContainer
Definition AMReX_FabArrayBase.H:219
int localindex(int K) const noexcept
Return local index in the vector of FABs.
Definition AMReX_FabArrayBase.H:119
const DistributionMapping & DistributionMap() const noexcept
Return constant reference to associated DistributionMapping.
Definition AMReX_FabArrayBase.H:130
int local_size() const noexcept
Return the number of local FABs in the FabArray.
Definition AMReX_FabArrayBase.H:113
CpOp
parallel copy or add
Definition AMReX_FabArrayBase.H:393
const BoxArray & boxArray() const noexcept
Return a constant reference to the BoxArray that defines the valid region associated with this FabArr...
Definition AMReX_FabArrayBase.H:95
An Array of FortranArrayBox(FAB)-like Objects.
Definition AMReX_FabArray.H:344
typename std::conditional_t< IsBaseFab< FAB >::value, FAB, FABType >::value_type value_type
Definition AMReX_FabArray.H:355
void CMD_remote_setVal_gpu(value_type x, const CommMetaData &thecmd, int scomp, int ncomp)
Definition AMReX_FBI.H:696
void FB_local_add_cpu(const FB &TheFB, int scomp, int ncomp)
Definition AMReX_FBI.H:455
void FB_local_add_gpu(const FB &TheFB, int scomp, int ncomp, bool deterministic)
Definition AMReX_FBI.H:618
void FB_local_copy_gpu(const FB &TheFB, int scomp, int ncomp)
Definition AMReX_FBI.H:546
void CMD_local_setVal_gpu(value_type x, const CommMetaData &thecmd, int scomp, int ncomp)
Definition AMReX_FBI.H:666
Array4< typename FabArray< FAB >::value_type const > array(const MFIter &mfi) const noexcept
Definition AMReX_FabArray.H:561
void FB_local_copy_cpu(const FB &TheFB, int scomp, int ncomp)
Definition AMReX_FBI.H:396
TagVector< Array4CopyTag< value_type > > const * FB_get_local_copy_tag_vector(const FB &TheFB)
Definition AMReX_FBI.H:503
FAB & atLocalIdx(int L) noexcept
Return a reference to the FAB associated with local index L.
Definition AMReX_FabArray.H:532
a one-thingy-per-box distributed object
Definition AMReX_LayoutData.H:13
Iterator for looping ever tiles and boxes of amrex::FabArray based containers.
Definition AMReX_MFIter.H:88
bool isValid() const noexcept
Is the iterator valid i.e. is it associated with a FAB?
Definition AMReX_MFIter.H:172
This class is a thin wrapper around std::vector. Unlike vector, Vector::operator[] provides bound che...
Definition AMReX_Vector.H:29
Long size() const noexcept
Definition AMReX_Vector.H:54
__host__ __device__ Dim3 length(Array4< T > const &a) noexcept
Return the spatial extents of an Array4 in Dim3 form.
Definition AMReX_Array4.H:1378
__host__ __device__ Dim3 lbound(Array4< T > const &a) noexcept
Return the inclusive lower bounds of an Array4 in Dim3 form.
Definition AMReX_Array4.H:1350
int MyProc() noexcept
Definition AMReX_ParallelDescriptor.H:128
__host__ __device__ AMREX_FORCE_INLINE T Exch(T *address, T val) noexcept
Definition AMReX_GpuAtomic.H:487
__host__ __device__ AMREX_FORCE_INLINE T CAS(T *const address, T compare, T const val) noexcept
Definition AMReX_GpuAtomic.H:513
__host__ __device__ AMREX_FORCE_INLINE void AddNoRet(T *sum, T value) noexcept
Definition AMReX_GpuAtomic.H:283
__host__ __device__ AMREX_FORCE_INLINE T Add(T *sum, T value) noexcept
Definition AMReX_GpuAtomic.H:200
void copyAsync(HostToDevice, InIter begin, InIter end, OutIter result) noexcept
A host-to-device copy routine. Note this is just a wrapper around memcpy, so it assumes contiguous st...
Definition AMReX_GpuContainers.H:228
static constexpr HostToDevice hostToDevice
Definition AMReX_GpuContainers.H:105
void streamSynchronize() noexcept
Definition AMReX_GpuDevice.H:310
gpuStream_t gpuStream() noexcept
Definition AMReX_GpuDevice.H:291
Definition AMReX_Amr.cpp:50
__host__ __device__ void ignore_unused(const Ts &...)
No-op helper that marks variables as intentionally unused.
Definition AMReX.H:259
__host__ __device__ Array4< T > makeArray4(T *p, Box const &bx, int ncomp) noexcept
Definition AMReX_BaseFab.H:123
DistributionMapping const & DistributionMap(FabArrayBase const &fa)
Definition AMReX_FabArrayBase.cpp:2867
__host__ __device__ Dim3 begin(BoxND< dim > const &box) noexcept
Return the iterator begin coordinate of box as Dim3.
Definition AMReX_Box.H:2239
void ParallelFor(TypeList< CTOs... > ctos, std::array< int, sizeof...(CTOs)> const &runtime_options, T N, F &&f)
Definition AMReX_CTOParallelForImpl.H:202
BoxND< 3 > Box
Box is an alias for amrex::BoxND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:35
IntVectND< 3 > IntVect
IntVect is an alias for amrex::IntVectND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:38
void LoopConcurrentOnCpu(Dim3 lo, Dim3 hi, F const &f) noexcept
Definition AMReX_Loop.H:388
void Abort(const std::string &msg)
Print a fatal-error message to stderr and abort execution.
Definition AMReX.cpp:241
const int[]
Definition AMReX_BLProfiler.cpp:1664
__host__ __device__ Dim3 end(BoxND< dim > const &box) noexcept
Return the iterator end coordinate of box as Dim3.
Definition AMReX_Box.H:2257
BoxArray const & boxArray(FabArrayBase const &fa)
Definition AMReX_FabArrayBase.cpp:2862
__host__ __device__ constexpr int get(IntVectND< dim > const &iv) noexcept
Get I'th element of IntVectND<dim>
Definition AMReX_IntVect.H:1334
Definition AMReX_TagParallelFor.H:58
Definition AMReX_TagParallelFor.H:26
Definition AMReX_TagParallelFor.H:50
Array4< T > dfab
Definition AMReX_TagParallelFor.H:51
A multidimensional array accessor.
Definition AMReX_Array4.H:288
Definition AMReX_TagParallelFor.H:106
Definition AMReX_TagParallelFor.H:116
A simple struct holding 3 int values for a 3D index.
Definition AMReX_Dim3.H:24
int x
Definition AMReX_Dim3.H:24
Used by a bunch of routines when communicating via MPI.
Definition AMReX_FabArrayBase.H:194
Box sbox
Definition AMReX_FabArrayBase.H:196
int srcIndex
Definition AMReX_FabArrayBase.H:198
Box dbox
Definition AMReX_FabArrayBase.H:195
int dstIndex
Definition AMReX_FabArrayBase.H:197
FillBoundary.
Definition AMReX_FabArrayBase.H:487
Describes a source FAB region participating in a copy.
Definition AMReX_FBI.H:18
IntVect offset
Definition AMReX_FBI.H:21
Box dbox
Definition AMReX_FBI.H:20
FAB const * sfab
Definition AMReX_FBI.H:19
Definition AMReX_TypeTraits.H:51
Definition AMReX_TypeTraits.H:61
Definition AMReX_TypeTraits.H:277
Definition AMReX_TagParallelFor.H:156
Lightweight tag describing raw pointer copies (used for pack/unpack).
Definition AMReX_FBI.H:25
Box dbox
Definition AMReX_FBI.H:27
char const * p
Definition AMReX_FBI.H:26